code_begin = """if (!extension_loaded('curl')) {{
    print 'Curl Extension not found. Exiting';
    exit;
}}
$ch = curl_init();
curl_setopt($ch, CURLOPT_URL, '{url}');
// Set so curl_exec returns the result instead of outputting it.
curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
// Set verbosity
curl_setopt($ch, CURLOPT_VERBOSE, 1);
$headers = array();
"""


code_header = """
$headers[] = "{header}:{value}";
"""


code_proxy = """
curl_setopt($ch, CURLOPT_PROXY, '{proxy}');
"""


code_post = """
$content = "{data}";
curl_setopt($ch, CURLOPT_POST, 1);
curl_setopt($ch, CURLOPT_POSTFIELDS, $content);
"""


code_search = """
curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
$response = curl_exec($ch);

if (curl_errno($ch)) {{
 print curl_error($ch);
}} else {{
 curl_close($ch);
}}
print $response;
$string = "{search_string}";
// Blindly copied from http://stackoverflow.com/questions/10778318/test-if-a-string-is-regex
// Checks if the passed string is a regex or a simple string
if( preg_match("/^\/.+\/[a-z]*$/i",$string)) {{
    if (preg_match($string, $response, $match)) {{
        print 'Found a match!';
    }}
}}
else {{
    if (strpos($response,$string) !== false) {{
        print 'Found a match!';
    }}
}}
"""


code_nosearch = """
curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
$response = curl_exec($ch);

if (curl_errno($ch)) {
 print curl_error($ch);
} else {
 curl_close($ch);
}
print $response;
"""



code_begin = """require "typhoeus"

url = '{url}'

options = {{
    followlocation: true,
    verbose: true,
    method: :{method},
"""


code_proxy = """
    proxy: '{proxy}',
"""


code_header = """
    "{header}" => "{value}","""


code_headers = """
    headers: {{{headers}
    }},
"""


code_post = """
    body: "{data}"
"""


code_search = """
}}
req = Typhoeus::Request.new(url, options)
req.on_complete do |response|
  if response.success?
    puts 'Response #{{response.code}}:'
    begin
        require 'colorize'
        lib_available = true
    rescue LoadError
        lib_available = false
    end

    matched = response.body.match /{search_string}/

    original = response.body
    if matched then
        if lib_available then
            for i in 0..matched.length
                original.gsub! /#{{matched[i]}}/, "#{{matched[i]}}".green
            end
        else
            for i in 0..matched.length
                puts 'Matched item: #{{matched[i]}}'
            end
        end
    end
    puts original

  elsif response.timed_out?
    puts 'Request Timed Out!'
  elsif response.code == 0
    # Could not get an http response, something's wrong.
    puts response.return_message
  else
    # Received a non-successful http response.
    puts 'HTTP request failed: ' + response.code.to_s
  end
end

req.run
"""


code_nosearch = """
}
req = Typhoeus::Request.new(url, options)
req.on_complete do |response|
  if response.success?
    puts 'Response #{response.code}'
    puts response.body

  elsif response.timed_out?
    puts 'Request Timed Out!'
  elsif response.code == 0
    # Could not get an http response, something's wrong.
    puts response.return_message
  else
    # Received a non-successful http response.
    puts 'HTTP request failed: ' + response.code.to_s
  end
end

req.run
"""



code_begin = """#!/usr/bin/env bash
curl"""


code_header = """ --header "{header}:{value}" """


code_proxy = " -x {proxy}"


code_post = """ --data "{data}" """


code_search = """ | egrep --color " {search_string} |$" """


code_nosearch = """ -v --request {method} {url} {headers} --include"""



import sys


try:
    input = raw_input  # Python2/3 version
except NameError:
    pass


def callback_interactive():
    raw_request = []
    print("Enter raw request (press ^D to finish or ^C to quit)")
    while True:
        try:
            raw_request.append(input().strip())
        except (EOFError, KeyboardInterrupt):
            break
    return '\n'.join(raw_request).strip()


def callback_file(filepath):
    raw_request = ''
    try:
        raw_request = open(filepath).read()
    except (OSError, IOError) as e:
        sys.stderr.write("error: Failed to open '%s'\n\n" % filepath)
        raise e
    return raw_request


def callback_inline(raw_request):
    return raw_request


def callback_stdin():
    raw_request = ''
    while True:
        try:
            raw_request += input() + '\n'
        except EOFError:
            break
    return raw_request[:-1]


handlers = {
    'interactive': callback_interactive,
    'file': callback_file,
    'inline': callback_inline,
    'stdin': callback_stdin
}



"""

:synopsis: Define the basic script class that will generate the script code.

"""

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote
from importlib import import_module

from .url import get_url, check_valid_url


class AbstractScript(object):

    """Abstract representation of a script."""

    __language__ = ''

    code_begin = ''
    code_header = ''
    code_proxy = ''
    code_post = ''
    code_https = ''
    code_search = ''
    code_nosearch = ''

    def __init__(self, headers=None, details=None, search=None):
        """Initialize the script generation.

        :param list headers: Headers list containing fields like 'Host', 'User-Agent', etc.
        :param dict details: Request specific details dictionary like body and method of the request.
        :param str search: String to search for in the response to the request.

        :raises ValueError: When url is invalid.
        """
        self.load_attributes(self.__class__)
        self._script = ''
        self.headers = headers
        self.details = details
        self.search = search
        self.url = ''
        if self.details:
            self.url = self.encode_url(self.create_url())

    def generate_script(self, headers=None, details=None, search=None):
        """Generate script code.

        :param list headers: Headers list containing fields like 'Host', 'User-Agent', etc.
        :param dict details: Request specific details dictionary like body and method of the request.
        :param str search: String to search for in the response to the request.

        :raises ValueError: when unsupported HTTP method, invalid `headers` or `details` values.

        :return: Generated script code.
        :rtype: str
        """
        self.headers = headers or self.headers
        self.details = details or self.details
        self.search = search or self.search
        if not self.headers:
            raise ValueError("'headers' cannot be equal to '%s'" % self.headers)
        elif not self.details:
            raise ValueError("'details' cannot be equal to '%s'" % self.details)
        if not self.url and self.details:
            self.url = self.encode_url(self.create_url())
        if self.code_begin:
            self._script += self._generate_begin()
        if self.code_proxy:
            self._script += self._generate_proxy()
        method = self.details.get('method', '').strip().lower()
        if method == 'get':
            pass
        elif method == 'post':
            if self.code_post:
                self._script += self._generate_post()
        else:
            raise ValueError("'%s' is not supported! Only GET and POST are supported for now." % self.details['method'])
        if self.code_https:
            self._script += self._generate_https()
        self._script += self._generate_request()
        return self._script

    def _generate_begin(self):
        """Default generation of the beginning of the code.

        :return: Beginning of the code.
        :rtype: str
        """
        return self.code_begin

    def _generate_headers(self):
        """Default generation of request headers.

        :return: Code snippet with HTTP requests headers.
        :rtype: str
        """
        code = ''
        for item in self.headers:
            header, value = item.split(':', 1)
            code += self.code_header.format(header=header.replace('"', '\\"'), value=value.replace('"', '\\"'))
        return code

    def _generate_proxy(self):
        """Default generation of the proxy specific code.

        :return: Code snippet with the proxy information.
        :rtype: str
        """
        if 'proxy_host' in self.details and 'proxy_port' in self.details:
            return self.code_proxy.format(proxy='%s:%s' % (self.details['proxy_host'], self.details['proxy_port']))
        return ''

    def _generate_post(self):
        """Default generation of the post body code.

        :return: Code snippet containing body to be sent in request.
        :rtype: str
        """
        return self.code_post.format(data=self.details.get('data', '').replace('"', '\\"'))

    def _generate_https(self):
        """Default generation of the HTTPS specific code.

        :return: Code snippet with HTTPS setup.
        :rtype: str
        """
        return self.code_https

    def _generate_request(self):
        """Default generation of the request code.

        :return: Code snippet for the request to send.
        :rtype: str
        """
        code = ''
        if self.search:
            if self.code_search:
                code += self._generate_search(self.search)
        else:
            if self.code_nosearch:
                code += self._generate_nosearch()
        return code

    def _generate_search(self, search_string=''):
        """Default generation of the code having search functionality.

        :param str search_string: String to search for in the response to the request.

        :return: Code snippet with the HTTP response search feature.
        :rtype: str
        """
        return self.code_search.format(search_string=search_string.replace('"', '\\"'))

    def _generate_nosearch(self):
        """Default generation of the code having no search functionality.

        :return: Code snippet absent of HTTP response search feature.
        :rtype: str
        """
        return self.code_nosearch

    def create_url(self):
        """Create valid URL.

        :raises ValueError: When URL is invalid.

        :return: Created URL.
        :rtype: str
        """
        url = get_url(self.details.get('Host', ''), self.details.get('pre_scheme', '')) + self.details.get('path', '')
        if not check_valid_url(url):
            raise ValueError("Invalid URL '%s'." % url)
        return url

    def encode_url(self, url):
        """Check if the URL of the HTTP request needs encoding.

        :param str url: URL to encode if needed.

        :return: Encoded URL if encoding is needed.
        :rtype: str
        """
        http_verb_with_encoding = ['head', 'options', 'get']
        encoded_url = url
        if self.details.get('data') and (self.details.get('method', '').lower() in http_verb_with_encoding):
            encoded_url += quote(self.details['data'], '')
        return encoded_url

    @staticmethod
    def load_attributes(cls):
        """Loads attributes to Script class from a given script's template

        Imports the template file/module, assigns all the attributes defined in the template file to the given class.

        :param class cls: Script class to which template is to be loaded.

        :raises AttributeError: When __language__ attribute is not present.
        """
        templates_path = "{}.templates".format(__name__.split('.', 1)[0])
        if not hasattr(cls, '__language__'):
            raise AttributeError("__language__ not found in class: {}, attributes cannot be loaded".format(cls.__name__))
        template = import_module("{templates_path}.{class_template}".format(
            templates_path=templates_path,
            class_template=cls.__language__))
        attributes = (var for var in vars(template) if var.startswith('code_'))
        for attr in attributes:
            setattr(cls, attr, getattr(template, attr))



from __future__ import print_function

import sys
import argparse

from .interface import HttpRequestTranslator
from .input_handler import handlers


def init():
    args = take_args()
    hrt = process_args(args)
    codes = hrt.generate_code()
    if args.parse_args().beautify:
        for language in codes:
            print('=' * 5 + ' %s ' % language.upper() + '=' * 5)
            print(codes[language])
    else :
        print(''.join(v for v in codes.values()))


def take_args():
    """Entry point for the translator through CLI. Initializes parser using `argparse` library.

    :return:`argparse.ArgumentParser` instance.
    :rtype:class `argparse.ArgumentParser`
    """
    # TODO: use non-hardcoded list of supported languages.
    parser = argparse.ArgumentParser(
        description="Request Translator is a standalone tool that can translate "
                    "raw HTTP requests into bash/python/php/ruby scripts")
    request_group = parser.add_mutually_exclusive_group()
    parser.add_argument(
        "--language", "-l",
        action="append",
        help="Generates a script in language 'language' for given HTTP request. "
             "If you want to generate multiple scripts, separate the script's name with a <,>. "
             "Available languages: bash, php, python, ruby")
    parser.add_argument(
        "--beautify", "-bt",
        action="store_true",
        help="Beautify code printing when multiple languages are selected")
    parser.add_argument(
        "--proxy", "-p",
        nargs="?",
        const="127.0.0.1:8009",
        help="Generates command/script with relevant, specified proxy")
    parser.add_argument(
        "--search_string", "-ss",
        help="Sends the request and searches for the required string in the response (regex can be provided)")
    parser.add_argument(
        "--interactive", "-i",
        action="store_true",
        help="Interactive mode: read raw HTTP request from keyboard, hit enter when ready. Type 'Ctrl+D' or 'Ctrl+C'"
             "to exit from the interactive mode.")
    parser.add_argument(
        "--data", "-d",
        help="Add the data that you want to send along with the header")
    request_group.add_argument(
        "--request", "-r",
        help="Input the HTTP request")
    request_group.add_argument(
        "--file", "-f",
        help="Input file for HTTP request")
    request_group.add_argument(
        "--stdin", "-s",
        action="store_true",
        help="Enable stdin mode for HTTP request")
    return parser


def get_input_type(args):
    """Find input handler with its corresponding parameters based on CLI arguments.

    :param `argparse.Namespace` args: `argparse.Namespace` instance.

    :return: tuple of input type and corresponding options
    :rtype: tuple
    """
    input_type = None
    options = []

    if args.interactive:
        input_type = 'interactive'
    elif args.file:
        input_type = 'file'
        options.append(args.file)
    elif args.request:
        input_type = 'inline'
        options.append(args.request)
    elif args.stdin:
        input_type = 'stdin'

    return (input_type, options)


def get_input(input_type, *options):
    """Takes an input mode type and returns a raw request.

    :raises OSError, IOError: When file fails to open in FileInput mode.

    :return: raw request
    :rtype: str
    """
    if input_type in handlers:
        if options:
            return handlers[input_type](*options)
        else:
            return handlers[input_type]()
    return ''


def process_args(parser):
    """Process the arguments provided to the translator CLI and return a HTTPRequestTranslator object.

    .. note::

        Default language is set to 'bash'.

    :param class `argparse.ArgumentParser`: `argparse.ArgumentParser` instance.

    :raises ValueError: When proxy is invalid.
    :raises NoRequestProvided: When no request is provided.

    :return: HTTPRequestTranslator instance
    :rtype: `HTTPRequestTranslator`
    """
    args = parser.parse_args()
    argdict = vars(args)

    languages = ['bash']  # Default script language is set to bash.
    if argdict.get('language'):
        languages = map(lambda x: x.strip(), argdict['language'][0].split(','))

    input_type, options = get_input_type(args)
    if not input_type:
        parser.print_help()
        sys.exit(-1)

    raw_request = get_input(input_type, *options)

    hrt_obj = HttpRequestTranslator(
        request=raw_request,
        languages=languages,
        proxy=args.proxy,
        search_string=args.search_string,
        data=args.data)

    return hrt_obj



code_begin = """#!/usr/bin/python
from __future__ import print_function
import re
import pycurl
try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

def main():
    buffer = BytesIO()
    curl_handler = pycurl.Curl()
    curl_handler.setopt(curl_handler.URL, '{url}')
    curl_handler.setopt(curl_handler.WRITEDATA, buffer)
    curl_handler.setopt(curl_handler.HTTPHEADER, {headers})
    # for verbosity
    curl_handler.setopt(curl_handler.VERBOSE, True)
    # Follow redirects
    curl_handler.setopt(curl_handler.FOLLOWLOCATION, True)
    # For older PycURL versions:
    #curl_handler.setopt(curl_handler.WRITEFUNCTION, buffer.write)
"""


code_proxy = """
    curl_handler.setopt(curl_handler.PROXY, '{proxy}')
"""


code_post = """
    # Sets request method to POST
    curl_handler.setopt(curl_handler.POSTFIELDS, "{data}")  #expects body to urlencoded
"""


code_https = """
    curl_handler.setopt(pycurl.SSL_VERIFYPEER, 1)
    curl_handler.setopt(pycurl.SSL_VERIFYHOST, 2)
    # If providing updated certs
    # curl_handler.setopt(pycurl.CAINFO, "/path/to/updated-certificate-chain.crt")
"""


code_search = """
    try:
        curl_handler.perform()
    except pycurl.error, error:
        print('An error occurred: ', error)
    curl_handler.close()

    body = buffer.getvalue()
    # Body is a string on Python 2 and a byte string on Python 3.
    # If we know the encoding, we can always decode the body and
    # end up with a Unicode string.
    response = body.decode('iso-8859-1')

    match = re.findall(r"{search_string}", str(response))
    try:
        from termcolor import colored
        lib_available = True
    except ImportError:
        lib_available = False
    if match:
        for item in match:
            if lib_available:
                replace_string = colored(match[item], 'green')
                response = re.sub(match[item], replace_string, str(response))
            else:
                print("Matched item: ",item)

    print(response)


if __name__ == '__main__':
    main()
"""


code_nosearch = """
    try:
        curl_handler.perform()
    except pycurl.error, error:
        print('An error occurred: ', error)
    curl_handler.close()

    body = buffer.getvalue()
    # Body is a string on Python 2 and a byte string on Python 3.
    # If we know the encoding, we can always decode the body and
    # end up with a Unicode string.
    response = body.decode('iso-8859-1')

    print(response)


if __name__ == '__main__':
    main()
"""
